/********************************************************************
 *  Copyright (c) 2003-4, InfiniTec IT Soluions						*
 *																	*
 *  This program is distributed in the hope that					*
 *	it will be useful, but WITHOUT ANY WARRANTY; without even the	*	
 *	implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR *
 *	PURPOSE.														* 
 ********************************************************************/
using System;
using System.Collections;
using System.Threading;
using System.Diagnostics;

namespace InfiniTec.SystemUtilities
{
	/// <summary>
	/// Implements a class similar to the ThreadPool.
	/// Can be used in cases where the threadpool ist 
	/// inappropiate. Additionally, the number of processing
	/// threads can be adjusted.
	/// </summary>
	/// <remarks>
	/// 
	/// </remarks>
	/// <example>
	///	<code>
	/// ThreadPool pool;
	/// 
	/// public void Init() {
	///		pool = new ThreadPool(4);
	///		pool.ProcessObject += new ProcessObjectEventHandler(processObject);
	/// 
	///		pool.Enqueue(objectToProcess);
	/// }
	/// 
	/// void processObject(object source, ProcessObjectEventArgs e) {
	///		DoSomething(e.Object);
	///	}
	///	</code>
	/// </example>
	public class AdvancedThreadPool
	{
		private int _AbsoluteThreadCount;
		private Queue _WorkItemQueue;
		private Hashtable _RunningThreads;
		private Stack _IdleThreads;
		private TimeSpan _MaxIdleTime;
		private string _Name;
		private int _MaxThreads;

		/// <summary>
		/// This event is fired, when a thread is ready to process
		/// an object from the process queue.
		/// </summary>
		public event ProcessObjectEventHandler ProcessObject;

		void OnProcessObject(ProcessObjectEventArgs e) 
		{
			if (ProcessObject != null) ProcessObject(this, e);
		}

		/// <summary>
		/// Gets or sets the name for the threadpool threads.
		/// </summary>
		/// <remarks>Note: Running threads are not renamed when this value is changed during runtime.
		/// </remarks>
		public string Name
		{
			get
			{
				return _Name;
			}
			set
			{
				if (value == null) value = "";
				_Name = value;
			}
		}

		/// <summary>
		/// Gets or sets the maximum idle that a thread may wait in idle state before it is
		/// removed from the thread pool.
		/// </summary>
		public TimeSpan MaxIdleTime
		{
			get
			{
				return _MaxIdleTime;
			}
			set
			{
				if (0 >= value.TotalSeconds) throw new ArgumentException("MaxIdleTime must be greater 0");
				_MaxIdleTime = value;
			}
		}

		/// <summary>
		/// Gets or sets the maximum number of threads that are running on this instance of the threadpool.
		/// </summary>
		public int MaxThreads
		{
			get
			{
				return _MaxThreads;
			}
			set
			{
				if (_MaxThreads == value)
					return;

				if (value < 1) throw new ArgumentException("MaxThreads must be greater 1", "MaxThreads");
				_MaxThreads = value;
			}
		}

		/// <summary>
		/// Initializes an instance of this class with a thread
		/// count of five.
		/// </summary>
		public AdvancedThreadPool(): this(5)
		{
		}
		
		private void processorThread_ThreadStart()
		{
			Job job;
			bool timeout;
			ThreadStateInfo info;

			timeout = false;

			while (!timeout) 
			{
				if (_WorkItemQueue.Count == 0) 
				{
					// Return this thread to the Idle thread stack
					info = (ThreadStateInfo) _RunningThreads[Thread.CurrentThread];
					_RunningThreads.Remove(info.Thread);
					_IdleThreads.Push(info);

					timeout = !info.WaitHandle.WaitOne(MaxIdleTime, true);

					// If a new workitem is available, put this thread into the running thread
					// table.
					if (!timeout) _RunningThreads[info.Thread] = info;
				}

				// If a timeout has occured, quit the thread method.
				if (timeout) 
				{
					Trace.WriteLine(string.Format("Thread {0} quits due to timeout.", Thread.CurrentThread.Name));
					break;
				}

				// Get the workitem
				lock (_WorkItemQueue) 
				{
					if (_WorkItemQueue.Count > 0) job = (Job) _WorkItemQueue.Dequeue();
					else job = null;
				}

				if (job != null) 
				{
					// Process the workitem
					try 
					{
						if (job.Callback is ProcessObjectEventHandler) 
						{
							((ProcessObjectEventHandler) job.Callback)(this, new ProcessObjectEventArgs(job.State));
						}
						else job.Callback.DynamicInvoke((object[]) job.State);
					} 
					catch (Exception ex) 
					{
						Trace.WriteLine("Exception occured during execution of threadpool workitem: " + ex.ToString());
					}
				}
			}
		}

		/// <summary>
		/// Enqueues an object in the workitem queue.
		/// </summary>
		/// <param name="o">The object to be enqueued</param>
		public void Enqueue(object o) 
		{
			Enqueue(ProcessObject, o);
		}

		/// <summary>
		/// Enqueues the delegate along with its parameters into the workitem queue.
		/// </summary>
		/// <param name="callback">The delegate to be called during processing</param>
		/// <param name="args">The parameters of the delegate</param>
		public void Enqueue(Delegate callback, params object[] args) 
		{
			Enqueue(new Job(callback, args));
		}

		private void Enqueue(Job job) 
		{
			ThreadStateInfo info;

			lock (_WorkItemQueue) 
			{
				lock (_IdleThreads) 
				{
					// Put the workitem into the queue
					_WorkItemQueue.Enqueue(job);

					// Check whether an idle thread is available

					info = null;

					if (_IdleThreads.Count > 0)  
					{
						do 
						{
							// Get an idle thread and check whether it is alive or has starved.
							info = (ThreadStateInfo) _IdleThreads.Pop(); 
						} while ((_IdleThreads.Count > 0) && (!info.Thread.IsAlive));
						if (!info.Thread.IsAlive) info = null;
					}
						
					if (info == null) 
					{
						// If the maximum number of threads is already reached, return
						if ((_IdleThreads.Count + _RunningThreads.Count) >= MaxThreads) return;
						
						// Else create a new thread.
						info = new ThreadStateInfo(CreateNewThread());
					
						_RunningThreads[info.Thread] = info;
						// Notify the thread
					}
					info.WaitHandle.Set();
				}
			}
		}
		private Thread CreateNewThread() 
		{
			Thread thread;

			thread = new Thread(new ThreadStart(processorThread_ThreadStart));
			thread.Name = string.Format("{0} #{1}", Name, _AbsoluteThreadCount);
			thread.IsBackground = true;
			Trace.WriteLine("Creating new thread:" + thread.Name);

			thread.Start();

			_AbsoluteThreadCount++;

			return thread;
		}

		/// <summary>
		/// Initializes an instance of this class with the
		/// specified number of processing threads.
		/// </summary>
		/// <param name="maxThreads">The maximum number of threads used for processing
		/// the objects in the process queue.</param>
		public AdvancedThreadPool(int maxThreads) 
		{
			MaxIdleTime = TimeSpan.FromSeconds(30);
			Name = "AdvancedThreadPool thread";
			_MaxThreads = maxThreads;
			_AbsoluteThreadCount = 0;

			_IdleThreads = Stack.Synchronized(new Stack());
			_WorkItemQueue = Queue.Synchronized(new Queue());
			_RunningThreads = Hashtable.Synchronized(new Hashtable());
		}
	}
}